// SDK
#include "extension.h"
#include "shvector_handle.h"
#include <ISmmPlugin.h>
ISmmPlugin *mmplugin;
#include <sh_vector.h>
#include <sh_string.h>
using namespace SourceHook;
// Extension SDK stuff
SHVectorExtension g_SHVectorExtension;
SMEXT_LINK(&g_SHVectorExtension);

// Macro for ease of read
#define GET_VEC_CELL_MACRO \
	Handle_t hndl=static_cast<Handle_t>(params[1]);\
	HandleError err;\
	HandleSecurity sec;\
	sec.pOwner=NULL;\
	sec.pIdentity=myself->GetIdentity();\
	CVector<cell_t> *vec;\
	if ((err=g_pHandleSys->ReadHandle(hndl,g_SHVectorHandle,&sec,(void **)&vec))!=HandleError_None)\
	{\
		return pCtx->ThrowNativeError("Invalid file handle %x (error %d)", hndl, err);\
	}\
	if(!vec)\
	{\
		return pCtx->ThrowNativeError("Invalid handle %x (error %d)",hndl,err);\
	}
#define GET_VEC_STRING_MACRO \
	Handle_t hndl=static_cast<Handle_t>(params[1]);\
	HandleError err;\
	HandleSecurity sec;\
	sec.pOwner=NULL;\
	sec.pIdentity=myself->GetIdentity();\
	CVector<String> *vec;\
	if ((err=g_pHandleSys->ReadHandle(hndl,g_SHVectorHandle,&sec,(void **)&vec))!=HandleError_None)\
	{\
		return pCtx->ThrowNativeError("Invalid file handle %x (error %d)", hndl, err);\
	}\
	if(!vec)\
	{\
		return pCtx->ThrowNativeError("Invalid handle %x (error %d)",hndl,err);\
	}
#define GET_VEC_CELL_MACRO2 \
	Handle_t hndl2=static_cast<Handle_t>(params[2]);\
	HandleError err2;\
	HandleSecurity sec2;\
	sec2.pOwner=NULL;\
	sec2.pIdentity=myself->GetIdentity();\
	CVector<cell_t> *vec2;\
	if ((err2=g_pHandleSys->ReadHandle(hndl2,g_SHVectorHandle,&sec2,(void **)&vec2))!=HandleError_None)\
	{\
		return pCtx->ThrowNativeError("Invalid file handle %x (error %d)", hndl2, err2);\
	}\
	if(!vec2)\
	{\
		return pCtx->ThrowNativeError("Invalid handle %x (error %d)",hndl2,err2);\
	}
#define GET_VEC_STRING_MACRO2 \
	Handle_t hndl2=static_cast<Handle_t>(params[2]);\
	HandleError err2;\
	HandleSecurity sec2;\
	sec2.pOwner=NULL;\
	sec2.pIdentity=myself->GetIdentity();\
	CVector<String> *vec2;\
	if ((err2=g_pHandleSys->ReadHandle(hndl2,g_SHVectorHandle,&sec2,(void **)&vec2))!=HandleError_None)\
	{\
		return pCtx->ThrowNativeError("Invalid file handle %x (error %d)", hndl, err);\
	}\
	if(!vec2)\
	{\
		return pCtx->ThrowNativeError("Invalid handle %x (error %d)",hndl,err);\
	}

bool SHVectorExtension::SDK_OnLoad(char *error, size_t err_max, bool late)
{
	g_pShareSys->AddNatives(myself,shvector_natives);
	g_SHVectorHandle=g_pHandleSys->CreateType("SHVector",&g_SHVectorHandler,0,NULL,NULL,myself->GetIdentity(),NULL);
	return true;
}

void SHVectorExtension::SDK_OnUnload()
{
}

bool SHVectorExtension::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
	mmplugin=this;
	if(mmplugin)
		return true;
	return false;
}

static cell_t SHVectorCreate(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorCreate(type);
	// We need to create a vector based on the params
	CVector<cell_t> *handle=new CVector<cell_t>();
	return g_pHandleSys->CreateHandle(g_SHVectorHandle,handle,pCtx->GetIdentity(),myself->GetIdentity(),NULL);
}

static cell_t SHVectorFree_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorFree_Cell(vector);
	// Free our vector from the memory
	GET_VEC_CELL_MACRO;
	delete vec;
	return 1;
}

static cell_t SHVectorFree_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorFree_String(vector);
	// Free our vector from the memory
	GET_VEC_STRING_MACRO;
	delete vec;
	return 1;
}

static cell_t SHVectorAt_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorAt_Int(vector,pos);
	// Get the thing at the pos and return it as an int
	GET_VEC_CELL_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorAt_Cell: Invalid position for vector passed.\n",(int)vec->size(),pos);
		return 0;
	}
	return vec->at(pos);
}

static cell_t SHVectorAt_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorAt_String(vector,pos,outbuf,maxlen);
	// Get the thing at the pos and return it as a string
	GET_VEC_STRING_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorAt_String: Invalid position for vector passed.\n");
		return 0;
	}
	pCtx->StringToLocal(params[3],params[4],vec->at(pos).c_str());
	return 1;
}

static cell_t SHVectorSetAt_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSetAt_Cell(vector,pos,value);
	// Set the thing at the pos to value
	GET_VEC_CELL_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSetAt_Cell: Invalid position for vector passed.\n");
		return 0;
	}
	vec->at(pos)=params[3];
	return 1;
}

static cell_t SHVectorSetAt_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSetAt_String(vector,pos,value);
	// Set the thing at the pos to value
	GET_VEC_STRING_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSetAt_String: Invalid position for vector passed.\n");
		return 0;
	}
	char *str;
	pCtx->LocalToString(params[3],&str);
	vec->at(pos).assign(str);
	return 1;
}

static cell_t SHVectorClear_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorClear_Cell(vector);
	// Clear the vector
	GET_VEC_CELL_MACRO;
	vec->clear();
	return 1;
}

static cell_t SHVectorClear_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorClear_String(vector);
	// Clear the vector
	GET_VEC_STRING_MACRO;
	vec->clear();
	return 1;
}

static cell_t SHVectorIsEmpty_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorIsEmpty_Cell(vector);
	// Is the vector empty?
	GET_VEC_CELL_MACRO;
	return vec->empty();
}

static cell_t SHVectorIsEmpty_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorIsEmpty_Cell(vector);
	// Is the vector empty?
	GET_VEC_STRING_MACRO;
	return vec->empty();
}

static cell_t SHVectorSwap_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSwap_Cell(vector,pos1,pos2);
	// Swap the two contents of two indexes
	GET_VEC_CELL_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSwap_Cell: Invalid position for vector passed (1).\n");
		return 0;
	}
	int pos2=params[3];
	if(pos2>=(int)vec->size()||pos2<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSwap_Cell: Invalid position for vector passed (2).\n");
		return 0;
	}
	cell_t val1=vec->at(pos);
	cell_t val2=vec->at(pos2);
	vec->at(pos)=val2;
	vec->at(pos2)=val1;
	return 1;
}

static cell_t SHVectorSwap_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSwap_String(vector,pos1,pos2);
	// Swap the two contents of two indexes
	GET_VEC_STRING_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSwap_Cell: Invalid position for vector passed (1).\n");
		return 0;
	}
	int pos2=params[3];
	if(pos2>=(int)vec->size()||pos2<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorSwap_Cell: Invalid position for vector passed (2).\n");
		return 0;
	}
	String val1=vec->at(pos);
	String val2=vec->at(pos2);
	vec->at(pos)=val2;
	vec->at(pos2)=val1;
	return 1;
}

static cell_t SHVectorInsert_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorInsert_Cell(vector,where,what);
	// Inserts "what" into "vector" at "where"
	GET_VEC_CELL_MACRO;
	int place=params[2];
	if(place==0)
	{
		vec->insert(vec->begin(),params[3]);
		return 1;
	}
	else
	{
		vec->push_back(params[3]);
		return 1;
	}
	return 0;
}

static cell_t SHVectorInsert_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorInsert_String(vector,where,what);
	// Inserts "what" into "vector" at "where"
	GET_VEC_STRING_MACRO;
	char *str;
	pCtx->LocalToString(params[3],&str);
	int place=params[2];
	if(place==0)
	{
		vec->insert(vec->begin(),str);
		return 1;
	}
	else
	{
		vec->insert(vec->end(),str);
		return 1;
	}
	return 0;
}

static cell_t SHVectorErase_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorErase_Cell(vector,where);
	// Erases "where" from "vector"
	GET_VEC_CELL_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorErase_Cell: Invalid position for vector passed.\n");
		return 0;
	}
	vec->erase(vec->iterAt(pos));
	return 1;
}

static cell_t SHVectorErase_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorErase_String(vector,where);
	// Erases "where" from "vector"
	GET_VEC_STRING_MACRO;
	int pos=params[2];
	if(pos>=(int)vec->size()||pos<0)
	{
		pCtx->ThrowNativeError("**ERROR** SHVectorErase_String: Invalid position for vector passed.\n");
		return 0;
	}
	vec->erase(vec->iterAt(pos));
	return 1;
}

static cell_t SHVectorSize_Cell(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSize_Cell(vector);
	// How big is our vector?
	GET_VEC_CELL_MACRO;
	return vec->size();
}

static cell_t SHVectorSize_String(IPluginContext *pCtx, const cell_t *params)
{
	// SHVectorSize_String(vector);
	// How big is our vector?
	GET_VEC_STRING_MACRO;
	return vec->size();
}

const sp_nativeinfo_t shvector_natives[] = 
{
	{"SHVectorCreate",				SHVectorCreate},
	{"SHVectorAt_Cell",				SHVectorAt_Cell},
	{"SHVectorAt_String",			SHVectorAt_String},
	{"SHVectorSetAt_Cell",			SHVectorSetAt_Cell},
	{"SHVectorSetAt_String",		SHVectorSetAt_String},
	{"SHVectorClear_Cell",			SHVectorClear_Cell},
	{"SHVectorClear_String",		SHVectorClear_String},
	{"SHVectorIsEmpty_Cell",		SHVectorIsEmpty_Cell},
	{"SHVectorIsEmpty_String",		SHVectorIsEmpty_String},
	{"SHVectorSwap_Cell",			SHVectorSwap_Cell},
	{"SHVectorSwap_String",			SHVectorSwap_String},
	{"SHVectorInsert_Cell",			SHVectorInsert_Cell},
	{"SHVectorInsert_String",		SHVectorInsert_String},
	{"SHVectorErase_Cell",			SHVectorErase_Cell},
	{"SHVectorErase_String",		SHVectorErase_String},
	{"SHVectorSize_Cell",			SHVectorSize_Cell},
	{"SHVectorSize_String",			SHVectorSize_String},
	{"SHVectorFree_Cell",			SHVectorFree_Cell},
	{"SHVectorFree_String",			SHVectorFree_String},
	{NULL,							NULL},
};
